import cPickle as pickle
from anarci import anarci
import os
import sys
sys.path.insert(0,os.path.dirname(os.path.abspath(__file__)))
from aboss.region_definitions import Accept
from aboss.run_ANARCI_parsing import list_of_frames, good_aminos as amino_acids
from anarci.germlines import all_germlines as IMGT_germlines

a = Accept()
a.numbering_scheme = "imgt"
a.definition = "imgt"

def finding_V_frameworks(numbering, chain):
    """
    Here we IMGT number germline V genes
    """
    all_regions = {}
    for frame in list_of_frames[chain]:
        all_regions[frame] = {}
        a.set_regions([frame])
        for n in range(len(numbering)):
            if a.accept(numbering[n][0], chain) == 1: 
                if numbering[n][1] == '-':
                        continue
                else:
                        if numbering[n][0][1] != ' ':
                                all_regions[frame][str(numbering[n][0][0]) + numbering[n][0][1]] = numbering[n][1]					
                        else:
                                if frame == "cdrh3" or frame == "cdrl3":
                                        if numbering[n][0][0] > 110:
                                                continue
                                all_regions[frame][str(numbering[n][0][0])] = numbering[n][1]
    return all_regions

def finding_J_frameworks(numbering, chain):
    """
    Here we IMGT number germline V genes
    """
    all_regions = {}
    if chain == "H":
            list_of_frames = ["cdrh3","fwh4"] 
    elif chain == "L":
            list_of_frames = ["cdrl3","fwl4"] 
    for frame in list_of_frames:
        all_regions[frame] = {}
        a.set_regions([frame])
        for n in range(len(numbering)):
                if a.accept(numbering[n][0], chain) == 1: 
                        if numbering[n][1] == '-':
                                        continue
                        else:
                                if numbering[n][0][1] != ' ':
                                        all_regions[frame][str(numbering[n][0][0]) + numbering[n][0][1]] = numbering[n][1]					
                                else:
                                        if frame == "cdrh3" or frame == "cdrl3":
                                                if numbering[n][0][0] < 115:
                                                        continue
                                        all_regions[frame][str(numbering[n][0][0])] = numbering[n][1]
    return all_regions

def anarci_fun(list_of_inputs):
    """
    We use ANARCI to number IMGT germlines
    """
    function = []
    for n in range(len(list_of_inputs)):
            function.append(("scFv"+str(n), list_of_inputs[n]))
    output = anarci(function, scheme='imgt')
    return output

def remove_dashes(dictionary):
	return { gene : dictionary[gene].replace("-","") for gene in dictionary  } 

def adding_zeros(dictionary):
    amino_acids = sorted(list("QWERTYIPASDFGHKLCVNM"))
    for frame in dictionary:
            for pos in dictionary[frame]:
                    for amino in amino_acids:
                            if amino not in dictionary[frame][pos]:
                                    dictionary[frame][pos][amino] = 0
    return dictionary


class CreateIMGTReference(object):
    """class that takes species, antibody chain, and lists of observed V and J genes to create IMGT reference matrix
    Inputs:
            @species = species of antibody Ig-seq data
            @chain = Ig-seq data antibody chain
            @observed_V_gene = List of V genes found in the Ig-seq data
            @observed_J_gene = List of J genes found in the Ig-seq data

    Return:
            @chain_reference: IMGT reference
    """
    
    def __init__(self, species, chain,observed_V_gene, observed_J_gene):
        self.species = species
        self.chain = chain
        self.observed_V_gene = observed_V_gene
        self.observed_J_gene = observed_J_gene

        # output dictionary
        self.chain_reference = {}

        # Loading observed V and J genesm and also getting thier amino acid sequences
        self.list_of_V_germlines, self.list_of_J_germlines, self.IMGT_germline_V_species, self.IMGT_germline_J_species = self.loadVJgenes()

    def loadVJgenes(self):

        # Loading V and J genes seen in Ig-seq data 
        list_of_V_germlines = list(set([x for x in self.observed_V_gene if x != None]))
        list_of_J_germlines = list(set([x for x in self.observed_J_gene if x != None]))
    
        # Collecting amino acid sequences of IMGT genes seen in Ig-seq data 
        if self.chain == "H":
                IMGT_germline_V_species = remove_dashes(IMGT_germlines['V']['H'][self.species]) 
                IMGT_germline_J_species = remove_dashes(IMGT_germlines['J']['H'][self.species])
        elif self.chain == "L":
                IMGT_germline_V_species = {}
                IMGT_germline_J_species = {}
                IMGT_germline_V_lambda = remove_dashes(IMGT_germlines['V']['L'][self.species]) 
                IMGT_germline_J_lambda = remove_dashes(IMGT_germlines['J']['L'][self.species])
                IMGT_germline_V_kappa = remove_dashes(IMGT_germlines['V']['K'][self.species]) 
                IMGT_germline_J_kappa = remove_dashes(IMGT_germlines['J']['K'][self.species])
                IMGT_germline_V_species.update(IMGT_germline_V_lambda)
                IMGT_germline_V_species.update(IMGT_germline_V_kappa)
                IMGT_germline_J_species.update(IMGT_germline_J_lambda)
                IMGT_germline_J_species.update(IMGT_germline_J_kappa)	
        return list_of_V_germlines, list_of_J_germlines, IMGT_germline_V_species, IMGT_germline_J_species

    def getVGeneSequence(self, VgeneSeq):
        speciesdummyJGenes = {"H":{"human":'IGHJ1*01', "mouse":'IGHJ1*01', "rabbit":'IGHJ1*01', "pig":'IGHJ1*01', "rat":'IGHJ1*01',
                              "alpace":"IGHJ2*01",
                              "rhesus":"IGHJ5-1*01"},
                              "L":{ "human":"IGKJ1*01", "mouse":"IGKJ1*01", "rhesus":"IGKJ1*01", "pig":"IGKJ1*01", "rat":"IGKJ1*01",
                                  "rabbit":"IGKJ1-1*03"}}
        return VgeneSeq+"A"+self.IMGT_germline_J_species[speciesdummyJGenes[self.chain][self.species]]

    def getDummyVGene(self):
        
        speciesdummyVGenes = {"H":{
                            "human":"IGHV1-18*01", "mouse": 'IGHV1-12*01', "alpaca": "IGHV2S1*01",
                            "rabbit": "IGHV1S1*01", "rhesus": "IGHV4-2*01", "pig": "IGHV1-11*01", "rat":"IGHV2S18*01"},
                            "L":{
                            "human":"IGKV1-12*01", "mouse": 'IGLV1*01', "rat": 'IGKV10S6*01', "rabbit": 'IGKV1S2*01', "pig":'IGKV1-11*01'}
                            }
        return self.IMGT_germline_V_species[ speciesdummyVGenes[self.chain][self.species]]

    def collectingNumbering(self, output, finding_chain):

        numbering = output[0][0][0][0]  
        aa = finding_chain(numbering, self.chain)
        for framework in aa:
            if framework not in self.chain_reference:
                self.chain_reference[framework] = {} 
            for pos in aa[framework]:
                if pos not in self.chain_reference[framework]:
                        self.chain_reference[framework][pos] = {}
                for ele in aa[framework][pos]:
                    if ele not in self.chain_reference[framework][pos]:
                        self.chain_reference[framework][pos][ele] = 0
                    self.chain_reference[framework][pos][ele] += 1

    def reconstructAndNumber(self):
        # Here we iterate through V and J genes, and recreate a full-length ab sequence
        # which is required for ANARCI numbering                                      

        # V Gene
        for germ_seq in self.list_of_V_germlines:
                if germ_seq not in self.IMGT_germline_V_species:
                        continue
                gene_V_sequence = self.getVGeneSequence( self.IMGT_germline_V_species[germ_seq])

                output_V = anarci_fun([gene_V_sequence])
                if not output_V[0][0]:
                        continue
                self.collectingNumbering(output_V,finding_V_frameworks )

        # J Gene
        for germ_seq in self.list_of_J_germlines:
                if germ_seq not in self.IMGT_germline_J_species:
                        continue
                germ_J_sequence = self.IMGT_germline_J_species[germ_seq]

                most_common_V_germ_sequence = self.getDummyVGene()	

                complete_seq = most_common_V_germ_sequence + "A" + germ_J_sequence
                output_J = anarci_fun([complete_seq])
                if not output_J[0][0]:
                        continue
                self.collectingNumbering(output_J, finding_J_frameworks)

        self.chain_reference = adding_zeros(self.chain_reference)

        return self.chain_reference

if __name__ == '__main__':
	pass
