import os
import sys
import shutil

class QueredAmino():

    def __init__(self):
        self.dictionary = {}
    def addQuery(self,amino,qu):
        if amino not in self.dictionary:
                self.dictionary[amino] = []
        self.dictionary[amino].append(qu)

def check_tempCSV():
    if os.path.isdir("Temp_CSV"):
            shutil.rmtree("Temp_CSV")

def extractSequences(fh):

    chunk = ""
    chunk_start = False

    for line in fh:

        if line[:6] == "Query=":
            chunk_start = True
            chunk += line
            continue

        if chunk_start:
            chunk += line

        if line[:27] == "Effective search space used":
            yield chunk
            chunk = ""
            chunk_start = False

def parse(chunk):

    """
    We iterate a chunk and collect amino acid sequences
    """
    alignment = False
    aaseq, gene = "", ""
    queryname = ""
    segment = False
    domain_i = -1
    aln_i = 0
    segment_i= 0
    chunk = chunk.split("\n")
    prev_line = ""
    for i, line in enumerate(chunk):
        if 'Query=' in line:
            queryname = line.split('Query= ')[1]

            sys.stdout.write("\rParsing query %s" % queryname)

            sys.stdout.flush()
            continue
        line =  line.strip()
        line = line.split()
        if line:

            # Previous line to this is amino acid sequence
            if line[0].startswith("lcl") or "Query_" in line[0]:

                aaseq += "".join(prev_line)
        prev_line = line
    return queryname, aaseq

def openIgBlastn(fileName, threshold =False):
    """
    We open Igblastn outputs as a file object and then iterate over lines
    """
    trackedAmino = QueredAmino()
    val = 0
    index = 0
    for sequence in extractSequences(open(fileName)):
        index += 1
        if index == threshold:
            return trackedAmino.dictionary
        query,aa =  parse(sequence)
        if aa == "":
            continue
        trackedAmino.addQuery(aa, query)
        val += 1
    final_dict = trackedAmino.dictionary
    assert val == sum([len(final_dict[ss]) for ss in final_dict]), "The lengths are not matching"
    print "\n\tExtraction of sequences from Igblastn is finished\n\n\n\tThe ANARCI Parsing is initiated\n\n"
    return final_dict

def checkIgblastn(fileName):
    threshold = 1000
    seq_dict = openIgBlastn(fileName, threshold) 
    if seq_dict == {}:
        return False
    return True
