import cPickle as pickle
from anarci import run_anarci
from optparse import OptionParser
import os, copy
import multiprocessing as mp
import time
from os.path import join
import csv
import json,sys
sys.path.insert(0,os.path.dirname(os.path.abspath(__file__)))
from aboss.region_definitions import Accept
from aboss.ProgressBar import returnProgress 
csv.field_size_limit(sys.maxsize)

a = Accept()
a.numbering_scheme = "imgt"
a.definition = "imgt"

list_of_frames ={
        "H":["cdrh3","fwh1", "cdrh1", "fwh2", "cdrh2", "fwh3","fwh4"],
        "L":["cdrl3","fwl1", "cdrl1", "fwl2", "cdrl2", "fwl3","fwl4"]}

good_aminos = sorted(list("QWERTYIPASDFGHKLCVNM"))

CDR3 = {"H":"cdrh3","L":"cdrl3"}
FW1 = {"H":"fwh1","L":"fwl1"}
FW3 = {"H":"fwh3","L":"fwl3"}

def merge_dicts(*dict_args):
    result = {}
    for dictionary in dict_args:
        result.update(dictionary)
    return result

def rabbit_length_check(position_73,rabbit_position_84,regions,quality):
	if len(regions["fwh3"]) - position_73 + rabbit_position_84  < 38:   # Accounting for insertions at postion 73 and deletion at 84
		quality = ("Bad", "FW3 length is shorter IMGT defined"), rabbit_position_84
	if len(regions["fwh2"]) < 17:
		quality = ("Bad", "FW2 length is shorter IMGT defined")
	if len(regions["fwh4"]) < 10:
		quality = ("Bad", "FW4 length is shorter IMGT defined")
	if len(regions["cdrh3"]) > 37:   
		quality = ("Bad", "CDR-H3 is chimeric")
	return quality

def standard_species_length_check(position_73,regions,quality):

    # We iterate heavy chain antibody regions, and check for their IMGT defined lengths
    for region, length in [("fwh2",17), ("fwh3",38),("fwh4",11), ("cdrh3", 37)]:
        if region == "fwh3":
            if len(regions[region]) - position_73 < length:
                quality = ("Bad", "{0} length is shorter IMGT defined".format(region))
                return quality
        elif region == "cdrh3":
            if len(regions[region]) > length:
                quality = ("Bad", "CDR-H3 is chimeric")
                return quality
        else:
            if len(regions[region]) < length:
                quality = ("Bad", "{0} length is shorter IMGT defined".format(region))
                return quality
    return quality

def light_chain_species_length_check(absent_positions,position_73,regions,quality):
	if len(regions["fwl3"]) + absent_positions - position_73 < 38: # Accounting for insertions at postion 73,81,83
		quality = ("Bad", "FW3 length is shorter IMGT defined")
	if len(regions["fwl2"]) < 17:
		quality = ("Bad", "FW2 length is shorter IMGT defined")
	if len(regions["fwl4"]) < 10:
		quality = ("Bad", "FW4 length is shorter IMGT defined")
	return quality

def standard_species_length_check_fw1(fw,all_region_aa,quality):
	"""
	Here we check for indels in fw1. The problem is that for many Ig-seq datasets FW1 primers are used that start after position 15.
	"""
	frame1 =  [int(x) for x in all_region_aa[fw]] 
	if 10 not in frame1:
		if min(frame1) < 10:
			if max(frame1) - min(frame1) != len(frame1):
				quality = ("Bad", "FW1 length is shorter IMGT defined")
		else:
			if max(frame1) - min(frame1) + 1 != len(frame1):
				quality = ("Bad", "FW1 length is shorter IMGT defined")
	else:
		if max(frame1) - min(frame1) + 1 != len(frame1):
			quality = ("Bad", "FW1 length is shorter IMGT defined")
	return quality

def rabbit_length_check_fw1(fw,all_region_aa,rabbit_position_2,quality):
	"""
	Framework 1 sometimes misses position 2 in rabbits.
	"""
	frame1 =  [int(x) for x in all_region_aa[fw]]
	if 10 not in frame1:
		if min(frame1) == 1:
			if max(frame1) - min(frame1) - rabbit_position_2 != len(frame1):
				quality = ("Bad", "FW1 length is shorter IMGT defined")				
		elif min(frame1) >= 3 and min(frame1) < 10:
			if max(frame1) - min(frame1) != len(frame1):
				quality = ("Bad", "FW1 length is shorter IMGT defined")	
		else:
			if max(frame1) - min(frame1) + 1 != len(frame1):
				quality = ("Bad", "FW1 length is shorter IMGT defined")
	else:
		if max(frame1) - min(frame1) + 1 - rabbit_position_2 != len(frame1):
			quality = ("Bad", "FW1 length is shorter IMGT defined")
	return quality

def checkAminoViability(aa):
    if aa not in good_aminos and aa != "-":  # Checking here for unusual aa while skipping "-" entries at some position (e.g. cases of canonical CDRs)
        return False
    return True

def checkSkip(aa):
    if aa == "-":
        return True
    return False

def checking_structural_viability(numbering, species, chain):
	"""The main function of ANARCI parsing. Here we check for unusual indels in antibody regions across different species. We also check if the first numbered position of the amino acid sequence starts before postion 23
	Inputs:
		@numbering: a list of ANARCI numbered antibody residues
	Return:
		@all_region_aa: ANARCI parsed numbered residues that comply with out knowledge of antibody folding
		@quality: Sequence quality. Good == No unusual indels, 3 CDRs and fw4 are covered.
		@regions: Separate amino acid sequence into antibody regions
	"""
	all_region_aa = {}
	regions = {}
	quality = "good"
	Conserved_position = 0
	absent_positions = 2
	position_73 = 0
	if species == "rabbit":
		rabbit_position_2 = 1
		rabbit_position_84 = 1

        # Here we iterate over antibody regions
        # and check for structural viability
	for frame in list_of_frames[chain]:
		all_region_aa[frame] = {}
		region = ''
		a.set_regions([frame])
		for n in range(len(numbering)):
			if a.accept(numbering[n][0], chain) == 1:       
                                
                                # Checking for non-amino acid entries
                                if not checkAminoViability(numbering[n][1]):
                                    quality = ("Bad", "Weird amino acid")
                                    return None, quality, None

                                # Checking for missing amino acids
                                if checkSkip(numbering[n][1]):
                                    continue
                                # Here we check for indels outside canonical CDR
                                if numbering[n][0][1] != ' ':

                                    all_region_aa[frame][str(numbering[n][0][0]) + numbering[n][0][1]] = numbering[n][1]
                                    if frame != CDR3[chain]:    
                                            quality = ("Bad", str(numbering[n][0][0]) + numbering[n][0][1])
                                            return None,quality,None
                                else:
                                        all_region_aa[frame][str(numbering[n][0][0])] = numbering[n][1]

                                        # Here we record the key positions that must be present in the antibody sequence
                                        # We need to account for positions such as 73 that are often not present
                                        # In light chain, positions 81 and 82 are often absent
                                        # Position 118 tells us if numbering of CDR-H3 was successful

                                        if chain == "L":
                                                if frame == "fwl1":
                                                        if str(numbering[n][0][0]) == "23":
                                                                Conserved_position += 1
                                                if frame == "fwl3":
                                                        if str(numbering[n][0][0]) == "104":
                                                                Conserved_position += 1
                                                        if str(numbering[n][0][0]) == "81" or str(numbering[n][0][0]) == "82":
                                                                absent_positions -= 1
                                                        elif str(numbering[n][0][0]) == "73":
                                                                position_73 += 1
                                                elif frame == "fwl4":
                                                        if str(numbering[n][0][0]) == "118":
                                                                Conserved_position += 1
                                        elif chain == "H":
                                                if frame == "fwh1":
                                                        if str(numbering[n][0][0]) == "23":
                                                                Conserved_position += 1
                                                        elif str(numbering[n][0][0]) == "2" and species == "rabbit":
                                                                rabbit_position_2 -=1
                                                if frame == "fwh3":
                                                        if str(numbering[n][0][0]) == "104":
                                                                Conserved_position += 1
                                                        elif str(numbering[n][0][0]) == "73":
                                                                position_73 += 1
                                                        elif str(numbering[n][0][0]) == "84" and species == "rabbit":
                                                                rabbit_position_84 -=1
                                                elif frame == "fwh4":
                                                        if str(numbering[n][0][0]) == "118":
                                                                Conserved_position += 1
                                # Here we record amino acid sequences
                                # of the antibody regions
                                region += numbering[n][1]
		regions[frame] = region

        # if we do not have the key conserved positions, we stop       
	if Conserved_position != 3:
		quality = ("Bad", "No Conserved Position")
		return all_region_aa, quality, regions
        
        # As rabbit has different antibody region lengths to other analysed species (e.g. human, mouse, rat, alpaca)
        # Also Heavy and Light chains have different region lengths
	if chain == "H" and species != "rabbit":
		quality = standard_species_length_check(position_73,regions,quality)

	elif chain == "H" and species == "rabbit":
		quality = rabbit_length_check(position_73,rabbit_position_84,regions,quality)

	elif chain == "L":
		quality = light_chain_species_length_check(absent_positions,position_73,regions,quality)
	if quality == "good":   
            # Check for indels in framework 1 as some sequences do not start with position 1 
            fw = FW1[chain] 
            if species != "rabbit":
                quality = standard_species_length_check_fw1(fw,all_region_aa,quality)
            # Special case for rabbits
            else:
                quality = rabbit_length_check_fw1(fw,all_region_aa,rabbit_position_2,quality)
	return all_region_aa, quality, regions

def anarci_fun(list_of_inputs, ncores):
	"""ANARCI running function
		@list_of_inputs: List of antibody sequences
		@ncores: Number of CPU cores to engage
	Return:
		@output: ANARCI annotated and numbered sequences
	"""

	list_for_anarci_parsing = [("scFv"+str(n), list_of_inputs[n]) for n in range(len(list_of_inputs))]
	try:
		output = run_anarci(list_for_anarci_parsing, scheme='imgt', assign_germline = True, ncpu= ncores)
	except:
		print "ANARCI parsing failed to parse this chunk of sequences"			
	return output

def remove_stars(lis):
	return [x for x in lis if "*" not in x and len(x) < 152]  # We put length cut-off here

def check_V_and_J_gene_IMG_alignment(seq_id, species_matrix, gene, species):
	"""Function that iterates through ANARCI species matrix to find sequence identities
		@seq_id: Sequence ID in ANARCI numbered outputs
		@species_matrix: ANARCI top species alignements with corresponding scores
		@gene: v_gene or j_gene to retrive the best alignment scores

	Return:
		@IG_gene: IG gene, or None if no confident alignemnts
	"""
	IG_gene = False
	if species_matrix[seq_id][0]["germlines"][gene][0][0][0] != species:
		range_genes = len(species_matrix[seq_id][0]["germlines"][gene][0])
		for n in range(range_genes):
			if species_matrix[seq_id][0]["germlines"][gene][0][n][0] == species:
				if species_matrix[seq_id][0]["germlines"][gene][1][n] > 0.5:
					IG_gene = species_matrix[seq_id][0]["germlines"][gene][0][n][1]
				break
	else:
		if species_matrix[seq_id][0]["germlines"][gene][1][0] > 0.50:  
			IG_gene = species_matrix[seq_id][0]["germlines"][gene][0][0][1]
	return IG_gene

def check_IMGT_alignment(seq_id, species_matrix, species, chain):
	"""Checking that antibody sequences align to IGHV and IGHJ genes
		@seq_id: Sequence ID in ANARCI numbered outputs
		@species_matrix: ANARCI top species alignements with corresponding scores
	
	Return:
		@IG_V: IGHV gene
		@IG_J: IGHJ gene
		@IGH_IGJ_status: None if IGHV or IGHJ do not align to HMMs
	"""
	# Here we aling sequence to HMM 
	IGH_IGJ_status = "unknown"
	IG_V = check_V_and_J_gene_IMG_alignment(seq_id,species_matrix,"v_gene", species)
	IG_J = check_V_and_J_gene_IMG_alignment(seq_id,species_matrix,"j_gene", species)
	if chain == "H":	
		if not IG_J or not IG_V:
			IGH_IGJ_status = None
		elif IG_V == False or IG_J == False:
			IGH_IGJ_status = None
		elif IG_V[:3] != "IGH" or IG_J[:3] != "IGH":
			IGH_IGJ_status = None
	elif chain == "L":
		if not IG_J or not IG_V:
			IGH_IGJ_status = None
		elif IG_V[:3] == "IGH" or IG_J[:3] == "IGH":
			IGH_IGJ_status = None
		elif IG_V == False or IG_J == False:
			IGH_IGJ_status = None
	return IG_V, IG_J, IGH_IGJ_status

def checking_sequence_viability(ANARCIoutput, species_matrix, OriginalInputs, chain, species, SequenceTracker):
        """Function that initiates checking for indels, HMM alignmets,chimeric sequences
                @ANARCIoutput: ANARCI numbering outputs
                @species_matrix: ANARCI top species alignements with corresponding scores
                @OriginalInputs: Original full length antibody sequences
        Return:
                @ANARCI_parsed: ANARCI parsed sequences
                @correct_V_gene: Track of IG_V gene
                @correct_J_gene: Track of IG_J gene
                @ParsingOutputs: List of (Antibody sequence, Redundancy, IG_V, IG_J, Numbered residue/positions, Sequence ID)
        """
        ANARCI_parsed = {}
        correct_V_gene,correct_J_gene = {}, {}
        ParsingOutputs = []
        if ANARCIoutput == []:
            return ANARCI_parsed, correct_V_gene, correct_J_gene, ParsingOutputs
        for u in range(len(ANARCIoutput)):
                if not ANARCIoutput[u] or ANARCIoutput[u][0][0] == []:
                        continue
                else:
                        numbering = ANARCIoutput[u][0][0]
                        aa, status, regions = checking_structural_viability(numbering, species, chain)
                        if status != "good":
                            continue

                        # Some sequences have very long fw4 regions or duplicated fw4 regions
                        if len(OriginalInputs[u][OriginalInputs[u].rfind(regions[list_of_frames[chain][-1]]):]) > 13 or OriginalInputs[u].count(regions[list_of_frames[chain][-1]]) > 1:
                            continue
                        # Here we check for the correct antibody chain 
                        IG_V, IG_J, IGH_IGJ_status = check_IMGT_alignment(u, species_matrix, species, chain)

                        if IGH_IGJ_status == None:
                                continue

                        ANARCI_parsed[OriginalInputs[u]] = json.dumps(aa)
                        correct_V_gene[IG_V] = 1
                        correct_J_gene[IG_J] = 1

                        # Saving paring Outputs
                        ParsingOutputs.append((OriginalInputs[u], regions[CDR3[chain]],IG_V,IG_J,json.dumps(aa), SequenceTracker.dictionary[OriginalInputs[u]]))

        return ANARCI_parsed, correct_V_gene, correct_J_gene, ParsingOutputs

def write_to_tempcsv(to_csv_list):
	"""Function that writes temporary outputs to a csv file
	Inputs:
		@to_csv_list: List of (Antibody sequence, Redundancy, IG_V, IG_J, Numbered residue/positions, Sequence ID)
	"""
        with open(join("Temp_CSV","temporary.csv"), "ab") as csv_file:
            writer = csv.writer(csv_file, delimiter=',')
	    for entry in to_csv_list:	

                row = entry[0], entry[1],entry[2],entry[3],entry[4], entry[5]
                writer.writerow(row)

class TrackSequences():
	"""
	Here we keep track of Query names if Igblastn was the input  
	"""
	def __init__(self):
		self.dictionary = {}		
	def add_entry(self,aa,query):
		if type(query) == int:
			self.dictionary[aa] = "Missing"		
		else:
			self.dictionary[aa] = query
	def __len__(self):
		return len(self.dictionary)
	def reset(self):
		self.dictionary = {}

class Main(object):
    """Class that initiates ABOSS analysis.
            @final_dict: python dictionary,key is antibody sequence, values are corresponding redundancy
            @ncpu: number of cores used in aboss analysis
            @rate_of_analysis: number of sequences to be parsed by ANARCI per cycle. It prevents RAM overload
            @chain: H or L
            @species: human as default

    outputs:
            @count_correct: Total number of ANARCI parsed non-redundant sequences
            @correct_V_gene: V Gene sequences observed in Ig-seq data. These will be used to build the IMGT germline reference
            @correct_J_gene: J Gene sequences observed in Ig-seq data. These will be used to build the IMGT germline reference
            @regions_aaCount: Counts of amino acids at antibody positions in Ig-seq data
    """
    def __init__(self, final_dict, ncpu, rate_of_analysis, chain, species):

       self.final_dict = final_dict
       self.ncpu = ncpu
       self.rate_of_analysis = rate_of_analysis
       self.chain = chain
       self.species = species
       self.chunk_size = 1000
       self.createTempCSV()

    def createTempCSV(self):
        # Here we check if Temp_CSV exists
        # If does, we rewrite the content
	try:
            os.makedirs(join("Temp_CSV"))
	except:
            with open(join("Temp_CSV","temporary.csv"), "wb") as csv_file:
                writer = csv.writer(csv_file, delimiter=',')

    def printsequenceNumbs(self):

	print "The total number of non-redundant sequences is %d" %(len(self.final_dict))
	try:
		print "The total number of redundant sequences is %d" % (sum([ len(self.final_dict[ss]) for ss in self.final_dict  ]))
	except TypeError:
		print "The total number of redundant sequences is %d" % (sum(self.final_dict.values()))

    def checkRate(self, num_of_entries, count, rate_of_analysis):
        if num_of_entries - count <= rate_of_analysis:
            rate_of_analysis = num_of_entries - count
            if rate_of_analysis < 1:
                    return False
        return rate_of_analysis

    def run(self):

        # Here we run ANARCI Parsing

        # Taking time start
	t1 = time.time()

        # Dictionary that holds ANARCI parsed outputs
        # that we write to the CSV file
        # and then remove the content
	ANARCIparsed = {}

        # Dictionary that holds amino acid entries at positions
	self.regions_aaCount = {}
        
        # Here we record V and J genes previously seen in Ig-seq data
	correct_V_gene,correct_J_gene = {}, {}
	
        # A list that holds non-redundant sequences before I pass them to ANARCI
	inputs = []

        # Counters
	count = 0
	count_correct = 0

        # Total number of non-redundant sequences
	num_of_entries = len(self.final_dict)

        # Printing sequence numbers before
        self.printsequenceNumbs()

	self.SequenceTracker = TrackSequences()
	for nonRedundantSequence in self.final_dict:

                rate_of_analysis = self.checkRate(num_of_entries, count, self.rate_of_analysis)
                
		self.SequenceTracker.add_entry(nonRedundantSequence, self.final_dict[nonRedundantSequence])
		if len(self.SequenceTracker)>= rate_of_analysis: 
                        
                    # Incrementing count number    
                    count += rate_of_analysis
                    #print count 	
                    
                    # Removing entries that have "*" in them, or length greater than 152 amino acids
                    inputs = remove_stars(self.SequenceTracker.dictionary.keys())

                    output = anarci_fun(inputs, self.ncpu)

                    # ABOSS multiprocessing
                    pool = mp.Pool(processes=self.ncpu)
                    results = [ pool.apply_async(checking_sequence_viability, args = (output[1][i::self.ncpu], output[2][i::self.ncpu], inputs[i::self.ncpu], self.chain, self.species, self.SequenceTracker)) for i in xrange(self.ncpu) ]
                    correct_sequences_portion = [ r.get() for r in results ]
                    del results
                    pool.close()

                    # Reseting inputs
                    inputs = []
                    self.SequenceTracker.reset()
        
                    # Collecting outputs from multiprocessing
                    temp_write_to_csv_list = []
                    for portion in range(len(correct_sequences_portion)):				
                            ANARCIparsed = merge_dicts(correct_sequences_portion[portion][0], ANARCIparsed)
                            correct_V_gene = merge_dicts(correct_sequences_portion[portion][1], correct_V_gene)
                            correct_J_gene = merge_dicts(correct_sequences_portion[portion][2], correct_J_gene)
                            temp_write_to_csv_list += correct_sequences_portion[portion][3]
                    
                    # Write outputs to csv file
                    write_to_tempcsv(temp_write_to_csv_list)
                    temp_write_to_csv_list = []
                    del correct_sequences_portion
                    
                    # Saving outputs and printing status
                    if len(ANARCIparsed) >= self.chunk_size:  
                        self.aggregateResults(ANARCIparsed)
                        count_correct += len(ANARCIparsed)
                        ANARCIparsed = {}
                    progress = returnProgress(float(count)/num_of_entries, round((float(len(ANARCIparsed) + count_correct))/count,3))
                    sys.stdout.write(progress)
                    sys.stdout.flush()
    
        # Finishing Parsing and Printing residues at postions 23 and 104
        self.aggregateResults(ANARCIparsed)
	ANARCIparsed = {}
	del self.final_dict
        print self.regions_aaCount[FW3[self.chain]]["104"], ": Position 104"
        print self.regions_aaCount[FW1[self.chain]]["23"], ": Position 23"
	return count_correct, correct_V_gene,correct_J_gene, self.regions_aaCount

    def aggregateResults(self, ANARCIparsed):
        for sequence in ANARCIparsed:
            
            if type(self.final_dict[sequence]) == int:
                    redundancy = self.final_dict[sequence]
            else:
                    redundancy = len(self.final_dict[sequence])
            aa = json.loads(ANARCIparsed[sequence])
            for framework in aa:
                if framework not in self.regions_aaCount:
                    self.regions_aaCount[framework] = {}
                for pos in aa[framework]:
                    if pos not in self.regions_aaCount[framework]:
                            self.regions_aaCount[framework][pos] = {}
                    for ele in aa[framework][pos]:
                            if ele not in self.regions_aaCount[framework][pos]:
                                    self.regions_aaCount[framework][pos][ele] = 0
                            self.regions_aaCount[framework][pos][ele] += redundancy
if __name__ == '__main__':
	pass
